""" Utils for the test suite. """

import os


def is_alpine_linux():
    """Check if the system is running Alpine Linux."""
    if os.path.exists("/etc/os-release"):
        with open("/etc/os-release") as f:
            return any("ID=alpine" in line for line in f)
    return False



""" This module is used to check the security of the commands entered by the
user. It checks if the command is allowed, if the path is allowed, if the
command contains forbidden characters, etc.
"""

import sys
import re
import os
import shlex
import glob

# import lshell specifics
from lshell import utils


def warn_count(messagetype, command, conf, strict=None, ssh=None):
    """Update the warning_counter, log and display a warning to the user"""

    log = conf["logpath"]

    if ssh:
        return 1, conf

    if strict:
        conf["warning_counter"] -= 1
        if conf["warning_counter"] < 0:
            log.critical(f'*** forbidden {messagetype} -> "{command}"')
            log.critical("*** Kicked out")
            sys.exit(1)
        else:
            log.critical(f'*** forbidden {messagetype} -> "{command}"')
            sys.stderr.write(
                f"*** You have {conf['warning_counter']} warning(s) left,"
                " before getting kicked out.\n"
            )
            log.error(f"*** User warned, counter: {conf['warning_counter']}")
            sys.stderr.write("This incident has been reported.\n")
    elif not conf["quiet"]:
        log.critical(f"*** forbidden {messagetype}: {command}")

    # Return 1 to indicate a warning was triggered.
    return 1, conf


def tokenize_command(command):
    """Tokenize the command line into separate commands based on the operators"""

    try:
        lexer = shlex.shlex(command, posix=True)
        lexer.whitespace_split = True
        lexer.commenters = ""
        tokens = list(lexer)
    except ValueError:
        # Handle the exception and return an appropriate message or handle as needed
        return []
    return tokens


def expand_shell_wildcards(item):
    """Expand shell wildcards in the item and return the expanded path"""

    # Expand shell variables like $HOME
    item = os.path.expanduser(item)
    item = os.path.expandvars(item)
    item = os.path.realpath(item)  # this is a hack - needs to be reviewed
    # test if item is a directory
    expanded_items = glob.glob(item, recursive=True)
    if expanded_items:
        # Return all matches instead of just the first one
        item = expanded_items[0]

    return item


def check_path(line, conf, completion=None, ssh=None, strict=None):
    """Check if a path is entered in the line. If so, it checks if user
    are allowed to see this path. If user is not allowed, it calls
    warn_count. In case of completion, it only returns 0 or 1.
    """
    allowed_path_re = str(conf["path"][0])
    denied_path_re = str(conf["path"][1][:-1])

    line = tokenize_command(line)

    for item in line:
        tomatch = expand_shell_wildcards(item)
        if os.path.isdir(tomatch) and tomatch[-1] != "/":
            tomatch += "/"
        match_allowed = re.findall(allowed_path_re, tomatch)
        if denied_path_re:
            match_denied = re.findall(denied_path_re, tomatch)
        else:
            match_denied = None

        # if path not allowed
        # case path executed: warn, and return 1
        # case completion: return 1
        if not match_allowed or match_denied:
            if not completion:
                ret, conf = warn_count("path", tomatch, conf, strict=strict, ssh=ssh)
            return 1, conf

    if not completion:
        if not re.findall(allowed_path_re, os.getcwd() + "/"):
            ret, conf = warn_count("path", tomatch, conf, strict=strict, ssh=ssh)
            os.chdir(conf["home_path"])
            conf["promptprint"] = utils.updateprompt(os.getcwd(), conf)
            return 1, conf
    return 0, conf


def check_secure(line, conf, strict=None, ssh=None):
    """This method is used to check the content on the typed command.
    Its purpose is to forbid the user to user to override the lshell
    command restrictions.
    The forbidden characters are placed in the 'forbidden' variable.
    Feel free to update the list. Emptying it would be quite useless..: )

    A warning counter has been added, to kick out of lshell a user if he
    is warned more than X time (X being the 'warning_counter' variable).
    """

    # store original string
    oline = line

    # strip all spaces/tabs
    line = line.strip()

    # init return code
    returncode = 0

    # This logic is kept crudely simple on purpose.
    # At most we might match the same stanza twice
    # (for e.g. "'a'", 'a') but the converse would
    # require detecting single quotation stanzas
    # nested within double quotes and vice versa
    relist = re.findall(r"[^=]\"(.+)\"", line)
    relist2 = re.findall(r"[^=]\'(.+)\'", line)
    relist = relist + relist2
    for item in relist:
        if os.path.exists(item):
            ret_check_path, conf = check_path(item, conf, strict=strict)
            returncode += ret_check_path

    # parse command line for control characters, and warn user
    if re.findall(r"[\x01-\x1F\x7F]", oline):
        ret, conf = warn_count("control char", oline, conf, strict=strict, ssh=ssh)
        return ret, conf

    for item in conf["forbidden"]:
        # allow '&&' and '||' even if singles are forbidden
        if item in ["&", "|"]:
            if re.findall(rf"[^\{item}]\{item}[^\{item}]", line):
                ret, conf = warn_count("character", item, conf, strict=strict, ssh=ssh)
                return ret, conf
        else:
            if item in line:
                ret, conf = warn_count("character", item, conf, strict=strict, ssh=ssh)
                return ret, conf

    # check if the line contains $(foo) executions, and check them
    executions = re.findall(r"\$\([^)]+[)]", line)
    for item in executions:
        # recurse on check_path
        ret_check_path, conf = check_path(item[2:-1].strip(), conf, strict=strict)
        returncode += ret_check_path

        # recurse on check_secure
        ret_check_secure, conf = check_secure(item[2:-1].strip(), conf, strict=strict)
        returncode += ret_check_secure

    # check for executions using back quotes '`'
    executions = re.findall(r"\`[^`]+[`]", line)
    for item in executions:
        ret_check_secure, conf = check_secure(item[1:-1].strip(), conf, strict=strict)
        returncode += ret_check_secure

    # check if the line contains ${foo=bar}, and check them
    curly = re.findall(r"\$\{[^}]+[}]", line)
    for item in curly:
        # split to get variable only, and remove last character "}"
        if re.findall(r"=|\+|\?|\-", item):
            variable = re.split(r"=|\+|\?|\-", item, 1)
        else:
            variable = item
        ret_check_path, conf = check_path(variable[1][:-1], conf, strict=strict)
        returncode += ret_check_path

    # if unknown commands where found, return 1 and don't execute the line
    if returncode > 0:
        return 1, conf
    # in case the $(foo) or `foo` command passed the above tests
    elif line.startswith("$(") or line.startswith("`"):
        return 0, conf

    lines = utils.split_commands(line)

    for separate_line in lines:
        # remove trailing parenthesis
        separate_line = re.sub(r"\)$", "", separate_line)
        separate_line = " ".join(separate_line.split())
        splitcmd = separate_line.strip().split(" ")

        # Extract the command and its arguments
        command = splitcmd[0]
        command_args_list = splitcmd[1:]
        command_args_string = " ".join(command_args_list)
        full_command = f"{command} {command_args_string}".strip()

        # in case of a sudo command, check in sudo_commands list if allowed
        if command == "sudo" and command_args_list:
            # allow the -u (user) flag
            if command_args_list[0] == "-u" and command_args_list:
                sudocmd = command_args_list[2]
            else:
                sudocmd = command_args_list[0]
            if sudocmd not in conf["sudo_commands"] and command_args_list:
                ret, conf = warn_count(
                    "sudo command", oline, conf, strict=strict, ssh=ssh
                )
                return ret, conf

        # if over SSH, replaced allowed list with the one of overssh
        if ssh:
            conf["allowed"] = conf["overssh"]

        # # for all other commands check in allowed list
        # if command not in conf["allowed"] and command:
        #     ret, conf = warn_count("command", command, conf, strict=strict, ssh=ssh)
        #     return ret, conf

        # Check if the full command (with arguments) or just the command is allowed
        if (
            full_command not in conf["allowed"]
            and command not in conf["allowed"]
            and command
        ):
            ret, conf = warn_count("command", command, conf, strict=strict, ssh=ssh)
            return ret, conf

        # Check if the command contains any forbidden extensions
        if conf.get("allowed_file_extensions"):
            allowed_extensions = conf["allowed_file_extensions"]
            check_extensions, disallowed_extensions = check_allowed_file_extensions(
                full_command, allowed_extensions
            )
            if check_extensions is False:
                ret, conf = warn_count(
                    f"file extension {disallowed_extensions}",
                    full_command,
                    conf,
                    strict=strict,
                    ssh=ssh,
                )
                return ret, conf

    return 0, conf


def check_allowed_file_extensions(command_line, allowed_extensions):
    """Checks if any file extensions in the command line are allowed."""
    # Split the command using shlex to handle quotes and escape characters
    try:
        tokens = shlex.split(command_line)
    except ValueError as e:
        # Log error or provide user feedback on the invalid input
        print(f"Error parsing command line: {e}")
        return True, []

    # Extract file extensions from tokens
    extensions_in_command = []
    for token in tokens:
        match = re.search(r"\.\w+", token)
        if match:
            extensions_in_command.append(match.group())

    # Check each extension against the allowed_extensions list
    disallowed_extensions = [
        ext for ext in extensions_in_command if ext not in allowed_extensions
    ]

    # if len(disallowed_extensions) == 1:
    #     disallowed_extensions = disallowed_extensions[0]

    if disallowed_extensions:
        return False, disallowed_extensions
    return True, None



"""Functional tests for lshell"""

import os
import unittest
import inspect
from getpass import getuser
import time
import pexpect

# import lshell specifics
# pylint: disable=C0411
from test import test_utils

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"
LSHELL = f"{TOPDIR}/bin/lshell"
USER = getuser()
PROMPT = f"{USER}:~\\$"


class TestFunctions(unittest.TestCase):
    """Functional tests for lshell"""

    def do_exit(self, child):
        """Exit the shell"""
        child.sendline("exit")
        child.expect(pexpect.EOF)

    def test_51_grep_valid_log_entry(self):
        """F51 | Test that grep matches a valid log entry format."""
        pattern = (
            r"\[\d{2}/[A-Za-z]{3}/\d{4}:\d{2}:\d{2}:\d{2}\s+(?:-|\+)\d{4}\].+UID=[\w.]+"
        )
        f_name = inspect.currentframe().f_code.co_name
        log_file = f"{TOPDIR}/test/testfiles/{f_name}.log"
        command = f"grep -P '{pattern}' {log_file}"

        child = pexpect.spawn(
            (
                f"{LSHELL} --config {CONFIG} "
                f'--allowed "+ [\'grep\']" --forbidden "[]"'
            )
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8")
        self.assertIn("user123", output)
        self.do_exit(child)

    def test_52_grep_invalid_date_format(self):
        """F52 | Test that grep matches a valid log entry format."""
        pattern = (
            r"\[\d{2}/[A-Za-z]{3}/\d{4}:\d{2}:\d{2}:\d{2}\s+(?:-|\+)\d{4}\].+UID=[\w.]+"
        )
        f_name = inspect.currentframe().f_code.co_name
        log_file = f"{TOPDIR}/test/testfiles/{f_name}.log"

        if test_utils.is_alpine_linux():
            command = f"grep -E '{pattern}' {log_file}"
        else:
            command = f"grep -P '{pattern}' {log_file}"

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} " '--allowed "+ [\'grep\']" --forbidden "[]"'
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8")
        self.assertNotIn("user123", output)
        self.do_exit(child)

    def test_53_grep_missing_uid(self):
        """F53 | Test that grep matches a valid log entry format."""
        pattern = (
            r"\[\d{2}/[A-Za-z]{3}/\d{4}:\d{2}:\d{2}:\d{2}\s+(?:-|\+)\d{4}\].+UID=[\w.]+"
        )
        f_name = inspect.currentframe().f_code.co_name
        log_file = f"{TOPDIR}/test/testfiles/{f_name}.log"

        if test_utils.is_alpine_linux():
            command = f"grep -E '{pattern}' {log_file}"
        else:
            command = f"grep -P '{pattern}' {log_file}"

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} " '--allowed "+ [\'grep\']" --forbidden "[]"'
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8")
        self.assertNotIn("user123", output)
        self.do_exit(child)

    def test_54_grep_special_characters_in_uid(self):
        """F54 | Test that grep matches a valid log entry format."""
        pattern = (
            r"\[\d{2}/[A-Za-z]{3}/\d{4}:\d{2}:\d{2}:\d{2}\s+(?:-|\+)\d{4}\].+UID=[\w.]+"
        )
        f_name = inspect.currentframe().f_code.co_name
        log_file = f"{TOPDIR}/test/testfiles/{f_name}.log"
        command = f"grep -P '{pattern}' {log_file}"

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} " '--allowed "+ [\'grep\']" --forbidden "[]"'
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8")
        self.assertIn("user.name", output)
        self.do_exit(child)

    def test_55_allowed_all_minus_list(self):
        """F55 | allow all commands minus the list"""

        command = "echo 1"
        expected = "*** forbidden command: echo"

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} " '--allowed \'"all" - ["echo"]'
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n")[1].strip()
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_56_path_minus_specific_path(self):
        """F56 | allow paths except for the specified path"""

        command1 = "cd /usr/"
        expected1 = f"{USER}:/usr\\$"
        command2 = "cd /usr/local"
        expected2 = "*** forbidden path: /usr/local/"

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} "
            '--path \'["/var", "/usr"] - ["/usr/local"]\''
        )
        child.expect(PROMPT)

        child.sendline(command1)
        child.expect(expected1)
        child.sendline(command2)
        child.expect(expected1)
        output = child.before.decode("utf-8").split("\n")[1].strip()
        self.assertEqual(expected2, output)
        self.do_exit(child)

    def test_57_overssh_all_minus_list(self):
        """F57 | overssh minus command list"""
        command = "echo 1"
        expected = (
            '*** forbidden char/command over SSH: "echo 1"\r\n'
            "This incident has been reported."
        )

        # add SSH_CLIENT to environment
        if not os.environ.get("SSH_CLIENT"):
            os.environ["SSH_CLIENT"] = "random"

        child = pexpect.spawn(
            f"{LSHELL} "
            f"--config {CONFIG} "
            f"--overssh \"['ls','echo'] - ['echo']\" "
            f"-c '{command}'"
        )
        child.expect(pexpect.EOF)

        output = child.before.decode("utf-8").strip()
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_58_allowed_plus_minus_list(self):
        """F58 | allow plus list minus list"""
        command = "echo 1"
        expected = "*** forbidden command: echo"

        child = pexpect.spawn(
            f"{LSHELL} "
            f"--config {CONFIG} "
            f"--allowed \"['ls'] + ['echo'] - ['echo']\" "
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n")[1].strip()
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_59a_forbidden_remove_one(self):
        """F59a | remove all items from forbidden list"""

        command = "echo 1 ; echo 2"
        expected = [" echo 1 ; echo 2\r", "1\r", "2\r", ""]

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} " '--forbidden \'[";"] - [";"]\''
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n")
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_59b_forbidden_remove_one(self):
        """F59b | fixed forbidden list"""

        command = "echo 1 ; echo 2"
        expected = "*** forbidden character: ;"

        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} " "--forbidden '[\";\"]'")
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n")[1].strip()
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_59c_forbidden_remove_one(self):
        """F59c | remove an item from forbidden list"""

        command = "echo 1 ; echo 2"
        expected = [" echo 1 ; echo 2\r", "1\r", "2\r", ""]

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} " '--forbidden \'[";", "|", "%"] - [";"]\''
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n")
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_60_allowed_extension_success(self):
        """F60 | allow extension and cat file with similar extension"""

        f_name = inspect.currentframe().f_code.co_name
        log_file = f"{TOPDIR}/test/testfiles/{f_name}.log"
        command = f"cat {log_file}"
        expected = "Hello world!"

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} "
            "--allowed \"+ ['cat']\" "
            "--allowed_file_extensions \"['.log']\""
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n")[1].strip()
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_61_allowed_extension_fail(self):
        """F61 | allow extension and cat file with different extension"""

        command = f"cat {CONFIG}"
        expected = f"*** forbidden file extension ['.conf']: cat {CONFIG}"

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} "
            "--allowed \"+ ['cat']\" "
            "--allowed_file_extensions \"['.log']\""
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n")[1].strip()
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_62_allowed_extension_empty(self):
        """F62 | allow extension empty and cat any file extension"""

        command = f"cat {CONFIG}"
        expected = "[global]"

        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} "
            "--allowed \"+ ['cat']\" "
            '--allowed_file_extensions "[]"'
        )
        child.expect(PROMPT)

        child.sendline(command)
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n")[1].strip()
        self.assertEqual(expected, output)
        self.do_exit(child)

    def test_63_multi_line_command(self):
        """F63 | Test multi-line command execution using line continuation"""

        # Start the shell process with lshell config
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} ")
        child.expect(PROMPT)

        # Send a multi-line command using line continuation
        child.sendline("echo 'First line' \\")
        child.sendline("'and second line'")
        child.expect(PROMPT)

        output = child.before.decode("utf-8").split("\n")[2].strip()
        expected_output = "First line and second line"
        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

        # Send an exit command to end the shell session
        self.do_exit(child)

    def test_64_multi_line_command_with_two_echos(self):
        """F64 | Test multi-line command execution with two echo commands"""

        # Start the shell process with lshell config
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --forbidden \"-[';']\"")
        child.expect(PROMPT)

        # Send two echo commands on two lines
        child.sendline("echo 'First line'; echo \\")
        child.sendline("'Second line';")
        child.expect(PROMPT)

        output = child.before.decode("utf-8").split("\n")[2:4]
        expected_output = ["First line\r", "Second line\r"]
        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

        # Send an exit command to end the shell session
        self.do_exit(child)

    def test_65_multi_line_command_security_echo(self):
        """F65 | test help, then echo FREEDOM! && help () sh && help"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG}  --forbidden \"-[';']\"")
        child.expect(PROMPT)

        # Step 1: Enter `help` command
        expected_help_output = (
            "bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo"
        )
        child.sendline("help")
        child.expect(PROMPT)
        help_output = child.before.decode("utf8").split("\n", 2)[1].strip()

        self.assertEqual(expected_help_output, help_output)

        # Step 2: Enter `echo FREEDOM! && help () sh && help`
        expected_output = (
            "1\r\nFREEDOM!\r\n"
            "bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo\r\n"
            "bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo"
        )
        child.sendline("echo 1; \\")
        child.expect(">")
        child.sendline("echo FREEDOM! && help () sh && help")
        child.expect(PROMPT)

        result = child.before.decode("utf8").strip().split("\n", 1)[1]

        # Verify the combined output
        self.assertEqual(expected_output, result)
        self.do_exit(child)

    def test_66_multi_line_command_ctrl_c(self):
        """F66 | Test multi-line command then ctrl-c to cancel"""

        # Start the shell process with lshell config
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} ")
        child.expect(PROMPT)

        # Send a multi-line command using line continuation
        child.sendline("echo 1 \\")
        child.expect(">")
        child.sendcontrol("c")
        child.expect(PROMPT)

        output = child.before.decode("utf-8").split("\n")[1].strip()
        expected_output = ""
        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

        # Send an exit command to end the shell session
        self.do_exit(child)

    def test_67_unclosed_quotes_traceback(self):
        """F67 | Test that unclsed quotes do not cause a traceback"""

        # Start the shell process with lshell config
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} ")
        child.expect(PROMPT)

        # Send a multi-line command using line continuation
        child.sendline('echo "OK""')
        child.expect("> ")
        child.sendline('OK"')
        child.expect(PROMPT)

        output = child.before.decode("utf-8").split("\n")[1].strip()
        expected_output = "OKOK"
        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

        # Send an exit command to end the shell session
        self.do_exit(child)

    def test_68_source_nonexistent_file(self):
        """F68 | Test sourcing a nonexistent environment file shows an error"""

        # Define a nonexistent file path
        env_file = "does_not_exist"

        # Start lshell and attempt to source the nonexistent file
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['source']\"")
        child.expect(PROMPT)

        # Source the nonexistent file and check for an error message
        child.sendline(f"source {env_file}")
        child.expect(PROMPT)

        output = child.before.decode("utf-8").split("\n")[1].strip()
        expected_output = f"ERROR: Unable to read environment file: {env_file}"

        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

        # Clean up and end session
        self.do_exit(child)

    def test_69_source_valid_file(self):
        """F69 | Test sourcing a valid environment file sets variables"""

        # Write a sample environment file
        env_file = "random_test_env"
        with open(env_file, "w") as file:
            file.write("export TEST_VAR='test_value'\n")

        # Start lshell and source the environment file
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['source']\"")
        child.expect(PROMPT)

        # Source the file and check if the variable is set
        child.sendline(f"source {env_file}")
        child.expect(PROMPT)
        child.sendline("echo $TEST_VAR")
        child.expect(PROMPT)

        output = child.before.decode("utf-8").split("\n")[1].strip()
        expected_output = f"test_value"

        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

        # Clean up and end session
        self.do_exit(child)

    def test_70_source_overwrite_variable(self):
        """F70 | Test sourcing a file overwrites existing environment variables"""

        # Write a sample environment file
        env_file = "test_env_overwrite"
        with open(env_file, "w") as file:
            file.write("export TEST_VAR='new_value'\n")

        # Start lshell, set initial variable, and source file to overwrite it
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['source']\"")
        child.expect(PROMPT)

        # Set initial variable and source the file
        child.sendline("export TEST_VAR='initial_value'")
        child.expect(PROMPT)
        child.sendline(f"source {env_file}")
        child.expect(PROMPT)
        child.sendline("echo $TEST_VAR")
        child.expect(PROMPT)

        output = child.before.decode("utf-8").split("\n")[1].strip()
        expected_output = f"new_value"

        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

        # Clean up and end session
        self.do_exit(child)

    def test_71_backgrounding_with_ctrl_z(self):
        """F71 | est backgrounding a command with Ctrl+Z."""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['tail']\"")
        child.expect(PROMPT)

        # Start a long-running command
        for file in ["file1", "file2", "file3"]:
            with open(file, "w") as f:
                f.write(f"{file} content")
            child.sendline(f"tail -f {file}")
            time.sleep(1)
            child.sendcontrol("z")
            # Verify stopped job message
            child.expect(r"\[\d+\]\+  Stopped        tail -f", timeout=1)

        # Check jobs output
        child.expect(PROMPT)
        child.sendline("jobs")
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n", 1)[1].strip()
        expected_output = (
            "[1]   Stopped        tail -f file1\r\n"
            "[2]-  Stopped        tail -f file2\r\n"
            "[3]+  Stopped        tail -f file3"
        )

        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

        # Resume the stopped job
        child.sendline("fg")
        child.sendcontrol("c")
        child.sendline("fg")
        child.sendcontrol("c")
        child.sendline("fg")
        child.sendcontrol("c")
        child.expect(PROMPT)

    def test_72_background_command_with_ampersand(self):
        """F72 | Test backgrounding a command with `&`."""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['sleep']\"")
        child.expect(PROMPT)

        # Run a background command with &
        child.sendline("sleep 60 &")
        child.expect(r"\[\d+\] sleep 60 \(pid: \d+\)", timeout=5)
        child.sendline("sleep 60 &")
        child.expect(r"\[\d+\] sleep 60 \(pid: \d+\)", timeout=5)
        child.sendline("sleep 60 &")
        child.expect(r"\[\d+\] sleep 60 \(pid: \d+\)", timeout=5)

        # Verify it's listed in jobs
        child.expect(PROMPT)
        child.sendline("jobs")
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n", 1)[1].strip()
        expected_output = (
            "[1]   Stopped        sleep 60\r\n"
            "[2]-  Stopped        sleep 60\r\n"
            "[3]+  Stopped        sleep 60"
        )

        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"

    def test_73_exit_with_stopped_jobs(self):
        """F73 | Test exiting with stopped jobs."""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['tail']\"")
        child.expect(PROMPT)

        # Start a long-running command and background it
        child.sendline("tail -f")
        time.sleep(1)
        child.sendcontrol("z")
        child.expect(r"\[\d+\]\+  Stopped        tail -f", timeout=1)

        # Attempt to exit
        child.sendline("exit")
        child.expect("There are stopped jobs.", timeout=5)

        # Verify stopped jobs are listed
        child.sendline("jobs")
        child.expect(r"\[\d+\]\+  Stopped        tail -f", timeout=5)

        # Exit again
        child.sendline("exit")
        child.expect(pexpect.EOF, timeout=5)

    def test_74_resume_stopped_jobs(self):
        """F74 | Test resuming stopped jobs."""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['tail']\"")
        child.expect(PROMPT)

        # Start and stop multiple jobs
        child.sendline(f"tail -f")
        time.sleep(1)
        child.sendcontrol("z")
        child.expect(r"\[\d+\]\+  Stopped        tail -f", timeout=1)
        child.sendline(f"tail -ff")
        time.sleep(1)
        child.sendcontrol("z")
        child.expect(r"\[\d+\]\+  Stopped        tail -ff", timeout=1)
        child.sendline(f"tail -fff")
        time.sleep(1)
        child.sendcontrol("z")
        child.expect(r"\[\d+\]\+  Stopped        tail -fff", timeout=1)
        child.sendline(f"tail -ffff")
        time.sleep(1)
        child.sendcontrol("z")
        child.expect(r"\[\d+\]\+  Stopped        tail -ffff", timeout=1)

        # Resume the second job
        child.sendline("fg 2")
        child.expect("tail -ff", timeout=5)
        child.sendcontrol("c")  # Send Ctrl+C to stop the job
        child.expect(PROMPT)

        # Resume the first job
        child.sendline("fg 1")
        child.expect("tail -f", timeout=5)
        child.sendcontrol("c")  # Send Ctrl+C to stop the job
        child.expect(PROMPT)

        # Resume the last two jobs
        child.sendline("fg")
        child.expect("tail -ffff", timeout=5)
        child.sendcontrol("c")  # Send Ctrl+C to stop the job
        child.expect(PROMPT)
        child.sendline("fg")
        child.expect("tail -fff", timeout=5)
        child.sendcontrol("c")  # Send Ctrl+C to stop the job
        child.expect(PROMPT)

    def test_75_interrupt_background_commands(self):
        """F75 | Test that `Ctrl+C` does not interrupt background commands."""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['sleep']\"")
        child.expect(PROMPT)

        # Run a background command
        child.sendline("sleep 60 &")
        child.expect(r"\[\d+\] sleep 60 \(pid: \d+\)", timeout=5)

        # Interrupt the foreground process (should not affect background)
        child.sendcontrol("c")
        child.expect(PROMPT)

        # Verify the background command is still running
        child.sendline("jobs")
        child.expect(r"\[\d+\]\+  Stopped        sleep 60", timeout=5)

    def test_76_jobs_after_completion(self):
        """F76 | Test that completed jobs are removed from the `jobs` list."""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --allowed \"+['sleep']\"")
        child.expect(PROMPT)

        # Run a short-lived background command
        child.sendline("sleep 2 &")
        child.expect(r"\[\d+\] sleep 2 \(pid: \d+\)", timeout=5)

        # Wait for the process to complete
        time.sleep(3)

        # Verify jobs output is empty
        child.sendline("jobs")
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n", 1)[1].strip()
        assert output == "", f"Expected no jobs, got: '{output}'"

    def test_77_mix_background_and_foreground(self):
        """F77 | Test mixing background and foreground commands."""
        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} --allowed \"+['sleep', 'tail']\""
        )
        child.expect(PROMPT)

        # Start a background command
        child.sendline("sleep 60 &")
        child.expect(r"\[\d+\] sleep 60 \(pid: \d+\)", timeout=5)

        # Start and stop a foreground command
        child.sendline("tail -f file1")
        time.sleep(1)
        child.sendcontrol("z")
        child.expect(r"\[\d+\]\+  Stopped        tail -f", timeout=1)

        # Verify jobs output
        child.expect(PROMPT)
        child.sendline("jobs")
        child.expect(PROMPT)
        output = child.before.decode("utf-8").split("\n", 1)[1].strip()
        expected_output = (
            "[1]-  Stopped        sleep 60\r\n[2]+  Stopped        tail -f file1"
        )

        assert (
            output == expected_output
        ), f"Expected '{expected_output}', got '{output}'"



""" This file contains all the variables used in lshell """

import sys
import os

__version__ = "0.10.8"

# Required config variable list per user
required_config = ["allowed", "forbidden", "warning_counter"]

# set configuration file path depending on sys.exec_prefix
# on *Linux sys.exec_prefix = '/usr' and default path must be in '/etc'
# on *BSD sys.exec_prefix = '/usr/{pkg,local}/' and default path
# is '/usr/{pkg,local}/etc'
lshell_path = os.path.abspath(__file__).split("/lib/")[0]
if sys.exec_prefix != "/usr":
    # for *BSD
    CONF_PREFIX = sys.exec_prefix
elif lshell_path == "/home/ghantoos/.local":
    # for *Linux user install
    CONF_PREFIX = lshell_path
else:
    # for *Linux system-wide install
    CONF_PREFIX = ""
configfile = CONF_PREFIX + "/etc/lshell.conf"

# history file
HISTORY_FILE = ".lhistory"

# help text
USAGE = f"""Usage: lshell [OPTIONS]
  --config <file>   : Config file location (default {configfile})
  --<param> <value> : where <param> is *any* config file parameter
  -h, --help        : Show this help message
  --version         : Show version
"""

# Intro Text
INTRO = """You are in a limited shell.
Type '?' or 'help' to get the list of allowed commands"""

# configuration parameters
configparams = [
    "config=",
    "help",
    "version",
    "quiet=",
    "log=",
    "logpath=",
    "loglevel=",
    "logfilename=",
    "syslogname=",
    "allowed=",
    "allowed_file_extensions=",
    "forbidden=",
    "sudo_commands=",
    "warning_counter=",
    "aliases=",
    "intro=",
    "prompt=",
    "prompt_short=",
    "timer=",
    "path=",
    "home_path=",
    "env_path=",
    "allowed_cmd_path=",
    "env_vars=",
    "env_vars_files=",
    "scp=",
    "scp_upload=",
    "scp_download=",
    "sftp=",
    "overssh=",
    "strict=",
    "scpforce=",
    "history_size=",
    "history_file=",
    "path_noexec=",
    "allowed_shell_escape=",
    "winscp=",
    "disable_exit=",
    "include_dir=",
]

builtins_list = [
    "cd",
    "clear",
    "exit",
    "export",
    "history",
    "lpath",
    "lsudo",
    "help",
    "fg",
    "bg",
    "jobs",
]

FORBIDDEN_ENVIRON = (
    "LD_AOUT_LIBRARY_PATH",
    "LD_AOUT_PRELOAD",
    "LD_LIBRARY_PATH",
    "LD_PRELOAD",
    "LD_ORIGIN_PATH",
    "LD_DEBUG_OUTPUT",
    "LD_PROFILE",
    "GCONV_PATH",
    "HOSTALIASES",
    "LOCALDOMAIN",
    "LOCPATH",
    "MALLOC_TRACE",
    "NLSPATH",
    "RESOLV_HOST_CONF",
    "RES_OPTIONS",
    "TMPDIR",
    "TZDIR",
    "LD_USE_LOAD_BIAS",
    "LD_DEBUG",
    "LD_DYNAMIC_WEAK",
    "LD_SHOW_AUXV",
    "GETCONF_DIR",
    "LD_AUDIT",
    "NIS_PATH",
    "PATH",
)



""" Unit tests for lshell """

import os
import unittest
from getpass import getuser
from time import strftime, gmtime
from unittest.mock import patch

# import lshell specifics
from lshell.checkconfig import CheckConfig
from lshell.utils import get_aliases, updateprompt, parse_ps1, getpromptbase
from lshell.variables import builtins_list
from lshell import builtincmd
from lshell import sec

TOPDIR = f"{os.path.dirname(os.path.realpath(__file__))}/../"
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"


class TestFunctions(unittest.TestCase):
    """Unit tests for lshell"""

    args = [f"--config={CONFIG}", "--quiet=1"]
    userconf = CheckConfig(args).returnconf()

    def test_03_checksecure_doublepipe(self):
        """U03 | double pipes should be allowed, even if pipe is forbidden"""
        args = self.args + ["--forbidden=['|']"]
        userconf = CheckConfig(args).returnconf()
        input_command = "ls || ls"
        return self.assertEqual(sec.check_secure(input_command, userconf)[0], 0)

    def test_04_checksecure_forbiddenpipe(self):
        """U04 | forbid pipe, should return 1"""
        args = self.args + ["--forbidden=['|']"]
        userconf = CheckConfig(args).returnconf()
        input_command = "ls | ls"
        return self.assertEqual(sec.check_secure(input_command, userconf)[0], 1)

    def test_05_checksecure_forbiddenchar(self):
        """U05 | forbid character, should return 1"""
        args = self.args + ["--forbidden=['l']"]
        userconf = CheckConfig(args).returnconf()
        input_command = "ls"
        return self.assertEqual(sec.check_secure(input_command, userconf)[0], 1)

    def test_06_checksecure_sudo_command(self):
        """U06 | quoted text should not be forbidden"""
        input_command = "sudo ls"
        return self.assertEqual(sec.check_secure(input_command, self.userconf)[0], 1)

    def test_07_checksecure_notallowed_command(self):
        """U07 | forbidden command, should return 1"""
        args = self.args + ["--allowed=['ls']"]
        userconf = CheckConfig(args).returnconf()
        input_command = "ll"
        return self.assertEqual(sec.check_secure(input_command, userconf)[0], 1)

    def test_08_checkpath_notallowed_path(self):
        """U08 | forbidden command, should return 1"""
        args = self.args + ["--path=['/home', '/var']"]
        userconf = CheckConfig(args).returnconf()
        input_command = "cd /tmp"
        return self.assertEqual(sec.check_path(input_command, userconf)[0], 1)

    def test_09_checkpath_notallowed_path_completion(self):
        """U09 | forbidden command, should return 1"""
        args = self.args + ["--path=['/home', '/var']"]
        userconf = CheckConfig(args).returnconf()
        input_command = "cd /tmp/"
        return self.assertEqual(
            sec.check_path(input_command, userconf, completion=1)[0], 1
        )

    def test_10_checkpath_dollarparenthesis(self):
        """U10 | when $() is allowed, return 0 if path allowed"""
        args = self.args + ["--forbidden=[';', '&', '|','`','>','<', '${']"]
        userconf = CheckConfig(args).returnconf()
        input_command = "echo $(echo aze)"
        return self.assertEqual(sec.check_path(input_command, userconf)[0], 0)

    def test_11_checkconfig_configoverwrite(self):
        """U12 | forbid ';', then check_secure should return 1"""
        args = [f"--config={CONFIG}", "--strict=123"]
        userconf = CheckConfig(args).returnconf()
        return self.assertEqual(userconf["strict"], 123)

    def test_13_multiple_aliases_with_separator(self):
        """U13 | multiple aliases using &&, || and ; separators"""
        # enable &, | and ; characters
        aliases = {"foo": "foo -l", "bar": "open"}
        input_command = "foo; fooo  ;bar&&foo  &&   foo | bar||bar   ||     foo"
        return self.assertEqual(
            get_aliases(input_command, aliases),
            " foo -l; fooo  ; open&& foo -l  " "&& foo -l | open|| open   || foo -l",
        )

    def test_14_sudo_all_commands_expansion(self):
        """U14 | sudo_commands set to 'all' is equal to allowed variable"""
        args = self.args + ["--sudo_commands=all"]
        userconf = CheckConfig(args).returnconf()
        # exclude internal and sudo(8) commands
        exclude = builtins_list + ["sudo"]
        allowed = [x for x in userconf["allowed"] if x not in exclude]
        # sort lists to compare
        userconf["sudo_commands"].sort()
        allowed.sort()
        return self.assertEqual(allowed, userconf["sudo_commands"])

    def test_16_allowed_ld_preload_builtin(self):
        """U16 | builtin commands should NOT be prepended with LD_PRELOAD"""
        args = self.args + ["--allowed=['echo','export']"]
        userconf = CheckConfig(args).returnconf()
        # verify that export is not automatically added to the aliases (i.e.
        # prepended with LD_PRELOAD)
        return self.assertNotIn("export", userconf["aliases"])

    def test_17_allowed_exec_cmd(self):
        """U17 | allowed_shell_escape should NOT be prepended with LD_PRELOAD
        The command should not be added to the aliases variable
        """
        args = self.args + ["--allowed_shell_escape=['echo']"]
        userconf = CheckConfig(args).returnconf()
        # sort lists to compare
        return self.assertNotIn("echo", userconf["aliases"])

    def test_18_forbidden_environment(self):
        """U18 | unsafe environment are forbidden"""
        input_command = "export LD_PRELOAD=/lib64/ld-2.21.so"
        args = input_command
        retcode = builtincmd.export(args)[0]
        return self.assertEqual(retcode, 1)

    def test_19_allowed_environment(self):
        """U19 | other environment are accepted"""
        input_command = "export MY_PROJECT_VERSION=43"
        args = input_command
        retcode = builtincmd.export(args)[0]
        return self.assertEqual(retcode, 0)

    def test_20_winscp_allowed_commands(self):
        """U20 | when winscp is enabled, new allowed commands are automatically
        added (see man).
        """
        args = self.args + ["--allowed=[]", "--winscp=1"]
        userconf = CheckConfig(args).returnconf()
        # sort lists to compare, except 'export'
        exclude = list(set(builtins_list) - set(["export"]))
        expected = exclude + ["scp", "env", "pwd", "groups", "unset", "unalias"]
        expected.sort()
        allowed = userconf["allowed"]
        allowed.sort()
        return self.assertEqual(allowed, expected)

    def test_21_winscp_allowed_semicolon(self):
        """U21 | when winscp is enabled, use of semicolon is allowed"""
        args = self.args + ["--forbidden=[';']", "--winscp=1"]
        userconf = CheckConfig(args).returnconf()
        # sort lists to compare
        return self.assertNotIn(";", userconf["forbidden"])

    def test_22_prompt_short_0(self):
        """U22 | short_prompt = 0 should show dir compared to home dir"""
        expected = f"{getuser()}:~/foo$ "
        args = self.args + ["--prompt_short=0"]
        userconf = CheckConfig(args).returnconf()
        currentpath = f"{userconf['home_path']}/foo"
        prompt = updateprompt(currentpath, userconf)
        # sort lists to compare
        return self.assertEqual(prompt, expected)

    def test_23_prompt_short_1(self):
        """U23 | short_prompt = 1 should show only current dir"""
        expected = f"{getuser()}:foo$ "
        args = self.args + ["--prompt_short=1"]
        userconf = CheckConfig(args).returnconf()
        currentpath = f"{userconf['home_path']}/foo"
        prompt = updateprompt(currentpath, userconf)
        # sort lists to compare
        return self.assertEqual(prompt, expected)

    def test_24_prompt_short_2(self):
        """U24 | short_prompt = 2 should show full dir path"""
        expected = f"{getuser()}:{os.getcwd()}/foo$ "
        args = self.args + ["--prompt_short=2"]
        userconf = CheckConfig(args).returnconf()
        currentpath = f"{userconf['home_path']}/foo"
        prompt = updateprompt(currentpath, userconf)
        # sort lists to compare
        return self.assertEqual(prompt, expected)

    def test_25_disable_ld_preload(self):
        """U25 | empty path_noexec should disable LD_PRELOAD"""
        args = self.args + ["--allowed=['echo','export']", "--path_noexec=''"]
        userconf = CheckConfig(args).returnconf()
        # verify that no alias was created containing LD_PRELOAD
        return self.assertNotIn("echo", userconf["aliases"])

    def test_26_checksecure_quoted_command(self):
        """U26 | quoted command should be parsed"""
        input_command = 'echo 1 && "bash"'
        return self.assertEqual(sec.check_secure(input_command, self.userconf)[0], 1)

    def test_27_checksecure_quoted_command(self):
        """U27 | quoted command should be parsed"""
        input_command = '"bash" && echo 1'
        return self.assertEqual(sec.check_secure(input_command, self.userconf)[0], 1)

    def test_28_checksecure_quoted_command(self):
        """U28 | quoted command should be parsed"""
        input_command = "echo'/1.sh'"
        return self.assertEqual(sec.check_secure(input_command, self.userconf)[0], 1)

    def test_29_env_path_updates_path_variable(self):
        """U29 | Test that --env_path updates the PATH environment variable."""
        # store the original $PATH
        original_path = os.environ["PATH"]

        # Simulate passing the --env_path argument
        random_path = "/usr/random:/this_is_a_test"
        args = self.args + [
            f"--env_path='{random_path}'",
        ]
        CheckConfig(args).returnconf()

        # Verify that the $PATH has been updated correctly
        expected_path = f"{random_path}:{original_path}"

        # Assuming CheckConfig sets the environment variable
        self.assertEqual(os.environ["PATH"], expected_path)

        # Reset the PATH environment variable
        os.environ["PATH"] = original_path

    @patch("sys.exit")  # Mock sys.exit to prevent exiting the test on failure
    def test_30_invalid_new_path(self, mock_exit):
        """U30 | Test that an invalid new PATH triggers an error and sys.exit."""
        original_path = os.environ["PATH"]
        random_path = "/usr/random:/invalid$path"
        args = self.args + [
            f"--env_path='{random_path}'",
        ]

        # Simulate passing the --env_path argument
        CheckConfig(args).returnconf()

        # Check that sys.exit was called due to invalid path
        mock_exit.assert_called_once_with(1)

        # The PATH should not have been changed
        self.assertEqual(os.environ["PATH"], original_path)

    @patch("sys.exit")
    def test_31_new_path_starts_with_colon(self, mock_exit):
        """U31 | Test that a new PATH starting with a colon triggers an error."""
        original_path = os.environ["PATH"]
        random_path = ":/usr/random:/this_is_a_test"
        args = self.args + [
            f"--env_path='{random_path}'",
        ]

        # Simulate passing the --env_path argument
        CheckConfig(args).returnconf()

        # Check that sys.exit was called due to invalid path
        mock_exit.assert_called_once()

        # The PATH should not have been changed
        self.assertEqual(os.environ["PATH"], original_path)

    def test_32_lps1_user_host_time(self):
        r"""U32 | LPS1 using \u@\h - \t> format"""
        os.environ["LPS1"] = r"\u@\h - \t> "
        expected = f"{getuser()}@{os.uname()[1].split('.')[0]} - {strftime('%H:%M:%S', gmtime())}> "
        prompt = parse_ps1(os.getenv("LPS1"))
        self.assertEqual(prompt, expected)
        del os.environ["LPS1"]

    def test_33_lps1_with_cwd(self):
        r"""U33 | LPS1 should replace cwd with \w format"""
        os.environ["LPS1"] = r"\u:\w$ "
        expected = f"{getuser()}:{os.getcwd().replace(os.path.expanduser('~'), '~')}$ "
        prompt = parse_ps1(os.getenv("LPS1"))
        self.assertEqual(prompt, expected)
        del os.environ["LPS1"]

    def test_34_prompt_default_user_host(self):
        """U34 | Default config-based prompt should replace %u and %h"""
        userconf = CheckConfig(self.args).returnconf()
        userconf["prompt"] = "%u@%h"
        expected = f"{getuser()}@{os.uname()[1].split('.')[0]}"
        prompt = getpromptbase(userconf)
        self.assertEqual(prompt, expected)

    def test_35_updateprompt_lps1_defined(self):
        """U35 | LPS1 environment variable should override config-based prompt"""
        os.environ["LPS1"] = r"\u@\H \W$ "
        expected = f"{getuser()}@{os.uname()[1]} {os.path.basename(os.getcwd())}$ "
        userconf = CheckConfig(self.args).returnconf()
        prompt = updateprompt(os.getcwd(), userconf)
        self.assertEqual(prompt, expected)
        del os.environ["LPS1"]

    def test_36_updateprompt_home_path(self):
        """U36 | Prompt path should use '~' for home directory"""
        userconf = CheckConfig(self.args).returnconf()
        currentpath = userconf["home_path"]
        expected = f"{getuser()}:~$ "
        prompt = updateprompt(currentpath, userconf)
        self.assertEqual(prompt, expected)

    def test_37_updateprompt_short_prompt_level_1(self):
        """U37 | short_prompt = 1 should show only last directory in path"""
        userconf = CheckConfig(self.args).returnconf()
        userconf["prompt_short"] = 1
        currentpath = f"{userconf['home_path']}/foo/bar"
        expected = f"{getuser()}:bar$ "
        prompt = updateprompt(currentpath, userconf)
        self.assertEqual(prompt, expected)

    def test_38_updateprompt_short_prompt_level_2(self):
        """U38 | short_prompt = 2 should show full directory path"""
        userconf = CheckConfig(self.args).returnconf()
        userconf["prompt_short"] = 2
        currentpath = f"{userconf['home_path']}/foo/bar"
        expected = f"{getuser()}:{currentpath}$ "
        prompt = updateprompt(currentpath, userconf)
        self.assertEqual(prompt, expected)

    def test_39_updateprompt_path_inside_home(self):
        """U39 | Path inside home directory should start with '~'"""
        userconf = CheckConfig(self.args).returnconf()
        currentpath = f"{userconf['home_path']}/projects"
        expected = f"{getuser()}:~{currentpath[len(userconf['home_path']):]}$ "
        prompt = updateprompt(currentpath, userconf)
        self.assertEqual(prompt, expected)

    def test_40_updateprompt_absolute_path_outside_home(self):
        """U40 | Absolute path outside home should display fully in prompt"""
        userconf = CheckConfig(self.args).returnconf()
        currentpath = "/etc"
        expected = f"{getuser()}:{currentpath}$ "
        prompt = updateprompt(currentpath, userconf)
        self.assertEqual(prompt, expected)



""" Utils for lshell """

import re
import subprocess
import os
import sys
import random
import string
import shlex
from getpass import getuser
from time import strftime, gmtime
import signal

# import lshell specifics
from lshell import variables
from lshell import builtincmd


def usage(exitcode=1):
    """Prints the usage"""
    sys.stderr.write(variables.USAGE)
    sys.exit(exitcode)


def version():
    """Prints the version"""
    sys.stderr.write(f"lshell-{variables.__version__} - Limited Shell\n")
    sys.exit(0)


def random_string(length):
    """generate a random string"""
    randstring = ""
    for char in range(length):
        char = random.choice(string.ascii_letters + string.digits)
        randstring += char

    return randstring


def get_aliases(line, aliases):
    """Replace all configured aliases in the line"""

    for item in aliases.keys():
        escaped_item = re.escape(item)
        reg1 = rf"(^|;|&&|\|\||\|)\s*{escaped_item}([ ;&\|]+|$)(.*)"
        reg2 = rf"(^|;|&&|\|\||\|)\s*{escaped_item}([ ;&\|]+|$)"

        # in case alias begins with the same command
        # (this is until i find a proper regex solution..)
        aliaskey = random_string(10)

        while re.findall(reg1, line):
            (before, after, rest) = re.findall(reg1, line)[0]
            linesave = line

            line = re.sub(reg2, f"{before} {aliaskey}{after}", line, count=1)

            # if line does not change after sub, exit loop
            if linesave == line:
                break

        # replace the key by the actual alias
        line = line.replace(aliaskey, aliases[item])

    for char in [";"]:
        # remove all remaining double char
        line = line.replace(f"{char}{char}", f"{char}")
    return line


def split_commands(line):
    """Split the command line into separate commands based on the operators"""
    # in case ';', '|' or '&' are not forbidden, check if in line
    lines = []

    # Variable to track if we're inside quotes
    in_quotes = False

    # Starting position of the command segment
    if line[0] in ["&", "|", ";"]:
        start = 1
    else:
        start = 0

    # Iterate over the command line
    for i in range(1, len(line)):
        # Check for quotes to ignore splitting inside quoted strings
        if line[i] in ['"', "'"] and (i == 0 or line[i - 1] != "\\"):
            in_quotes = not in_quotes
            # Only split if we are not inside quotes and the current character
            # is an unescaped operator
        if line[i] in ["&", "|", ";"] and line[i - 1] != "\\" and not in_quotes:
            if start != i:
                lines.append(line[start:i])
            start = i + 1

    # Append the last segment of the command
    if start != len(line):
        lines.append(line[start:])

    return lines


def split_command_args(line):
    """Split the command line into cmd and args"""
    # Use shlex to split the command into parts
    tokens = shlex.split(line)

    if tokens:
        # The first token is the command
        cmd = tokens[0]
        # The rest are the arguments
        args = " ".join(tokens[1:])
    else:
        # If there are no tokens, return None for both
        cmd, args = "", ""

    return cmd, args


def replace_exit_code(line, retcode):
    """Replace the exit code in the command line. Replaces all occurrences of
    $? with the exit code."""
    if re.search(r"[;&\|]", line):
        p = re.compile(r"(\s|^)(\$\?)([\s|$]?[;&|].*)")
    else:
        p = re.compile(r"(\s|^)(\$\?)(\s|$)")

    line = p.sub(rf" {retcode} \3", line)

    return line


def cmd_parse_execute(command_line, shell_context=None):
    """Parse and execute a shell command line"""
    # Split command line by shell grammar: '&&', '||', and ';;'
    cmd_split = re.split(r"(;|&&|\|\|)", command_line)

    # Initialize return code
    retcode = 0

    # Iterate over commands and operators
    for i in range(0, len(cmd_split), 2):
        command = cmd_split[i].strip()
        operator = cmd_split[i - 1].strip() if i > 0 else None

        # Skip empty commands
        if not command:
            continue

        # Only execute commands based on the previous operator and return code
        if operator == "&&" and retcode != 0:
            continue
        elif operator == "||" and retcode == 0:
            continue

        # Get the executable command
        try:
            executable, argument = re.split(r"\s+", command, maxsplit=1)
        except ValueError:
            executable, argument = command, ""

        # Check if command is in built-ins list or execute it via exec_cmd
        if executable in variables.builtins_list:
            if executable == "help":
                shell_context.do_help(command)
            elif executable == "exit":
                shell_context.do_exit(command)
            elif executable == "history":
                builtincmd.history(shell_context.conf, shell_context.log)
            elif executable == "cd":
                retcode = builtincmd.cd(argument, shell_context.conf)
            else:
                retcode = getattr(builtincmd, executable)(shell_context.conf)
        else:
            if "path_noexec" in shell_context.conf:
                os.environ["LD_PRELOAD"] = shell_context.conf["path_noexec"]
            command = replace_exit_code(command, retcode)
            retcode = exec_cmd(command)

    return retcode


def exec_cmd(cmd):
    """Execute a command exactly as entered, with support for backgrounding via Ctrl+Z."""

    class CtrlZException(Exception):
        """Custom exception to handle Ctrl+Z (SIGTSTP)."""

        pass

    def handle_sigtstp(signum, frame):
        """Handle SIGTSTP (Ctrl+Z) by sending the process to the background."""
        if proc and proc.poll() is None:  # Ensure process is running
            proc.send_signal(signal.SIGSTOP)  # Stop the process
            builtincmd.background_jobs.append(proc)  # Add process to background jobs
            job_id = len(builtincmd.background_jobs)
            sys.stdout.write(f"\n[{job_id}]+  Stopped        {cmd}\n")
            sys.stdout.flush()
            raise CtrlZException()  # Raise custom exception for SIGTSTP handling

    def handle_sigcont(signum, frame):
        """Handle SIGCONT to resume a stopped job in the foreground."""
        if proc and proc.poll() is None:
            proc.send_signal(signal.SIGCONT)

    # Check if the command is to be run in the background
    background = cmd.strip().endswith("&")
    if background:
        # Remove '&' and strip any extra spaces
        cmd = cmd[:-1].strip()

    try:
        # Register SIGTSTP (Ctrl+Z) and SIGCONT (resume) signal handlers
        signal.signal(signal.SIGTSTP, handle_sigtstp)
        signal.signal(signal.SIGCONT, handle_sigcont)
        cmd_args = shlex.split(cmd)
        if background:
            with open(os.devnull, "r") as devnull_in, open(
                os.devnull, "w"
            ) as devnull_out:
                proc = subprocess.Popen(
                    cmd_args,
                    stdin=devnull_in,  # Redirect input to /dev/null
                    stdout=sys.stdout,
                    stderr=sys.stderr,
                    preexec_fn=os.setsid,
                )
            # add to background jobs and return
            builtincmd.background_jobs.append(proc)
            job_id = len(builtincmd.background_jobs)
            print(f"[{job_id}] {cmd} (pid: {proc.pid})")
            retcode = 0
        else:
            proc = subprocess.Popen(cmd_args, preexec_fn=os.setsid)
            proc.communicate()
            retcode = proc.returncode if proc.returncode is not None else 0

    except FileNotFoundError:
        sys.stderr.write(
            f"Command '{cmd_args[0]}' not found in $PATH or not installed on the system.\n"
        )
        retcode = 127
    except CtrlZException:  # Handle Ctrl+Z
        retcode = 0
    except KeyboardInterrupt:  # Handle Ctrl+C
        if proc and proc.poll() is None:
            os.killpg(os.getpgid(proc.pid), signal.SIGINT)
        retcode = 130

    return retcode


def parse_ps1(ps1):
    """Parse and format $PS1-style prompt with lshell-compatible values"""
    user = getuser()
    host = os.uname()[1]
    cwd = os.getcwd()
    home = os.path.expanduser("~")
    prompt_symbol = "#" if os.geteuid() == 0 else "$"

    # Define LPS1 replacement mappings
    replacements = {
        r"\u": user,
        r"\h": host.split(".")[0],
        r"\H": host,
        r"\w": cwd.replace(home, "~", 1) if cwd.startswith(home) else cwd,
        r"\W": os.path.basename(cwd),
        r"\$": prompt_symbol,
        r"\\": "\\",
        r"\t": strftime("%H:%M:%S", gmtime()),
        r"\T": strftime("%I:%M:%S", gmtime()),
        r"\A": strftime("%H:%M", gmtime()),
        r"\@": strftime("%I:%M:%S%p", gmtime()),
        r"\d": strftime("%a %b %d", gmtime()),
    }
    # Replace each placeholder with its corresponding value
    for placeholder, value in replacements.items():
        ps1 = ps1.replace(placeholder, value)

    return ps1


def getpromptbase(conf):
    """Get the base prompt structure, using $PS1 or defaulting to config-based prompt"""
    ps1_env = os.getenv("LPS1")
    if ps1_env:
        # Use $LPS1 with placeholders if defined
        promptbase = parse_ps1(ps1_env)
    else:
        # Fallback to configured prompt if no $PS1 is defined
        promptbase = conf.get("prompt", "%u")
        promptbase = promptbase.replace("%u", getuser())
        promptbase = promptbase.replace("%h", os.uname()[1].split(".")[0])

    return promptbase


def updateprompt(path, conf):
    """Set the prompt with updated path and user privilege level, supporting $LPS1 format"""
    promptbase = getpromptbase(conf)
    prompt_symbol = "# " if os.geteuid() == 0 else "$ "

    # Determine dynamic path display if $LPS1 is not defined
    if os.getenv("LPS1"):
        prompt = promptbase
    else:
        if path == conf["home_path"]:
            current_path = "~"
        elif conf.get("prompt_short") == 1:
            current_path = os.path.basename(path)
        elif conf.get("prompt_short") == 2:
            current_path = path
        elif path.startswith(conf["home_path"]):
            current_path = f"~{path[len(conf['home_path']):]}"
        else:
            current_path = path
        prompt = f"{promptbase}:{current_path}{prompt_symbol}"

    return prompt



""" This module contains the built-in commands of lshell """

import glob
import sys
import os
import readline
import signal

# import lshell specifics
from lshell import variables
from lshell import utils


# Store background jobs
background_jobs = []


def lpath(conf):
    """lists allowed and forbidden path"""
    if conf["path"][0]:
        sys.stdout.write("Allowed:\n")
        lpath_allowed = conf["path"][0].split("|")
        lpath_allowed.sort()
        for path in lpath_allowed:
            if path:
                sys.stdout.write(f" {path[:-2]}\n")
    if conf["path"][1]:
        sys.stdout.write("Denied:\n")
        lpath_denied = conf["path"][1].split("|")
        lpath_denied.sort()
        for path in lpath_denied:
            if path:
                sys.stdout.write(f" {path[:-2]}\n")
    return 0


def lsudo(conf):
    """lists allowed sudo commands"""
    if "sudo_commands" in conf and len(conf["sudo_commands"]) > 0:
        sys.stdout.write("Allowed sudo commands:\n")
        for command in conf["sudo_commands"]:
            sys.stdout.write(f" - {command}\n")
        return 0

    sys.stdout.write("No sudo commands allowed\n")
    return 1


def history(conf, log):
    """print the commands history"""
    try:
        try:
            readline.write_history_file(conf["history_file"])
        except IOError:
            log.error(f"WARN: couldn't write history to file {conf['history_file']}\n")
            return 1
        with open(conf["history_file"], "r", encoding="utf-8") as f:
            i = 1
            for item in f.readlines():
                sys.stdout.write(f"{i}:  {item}")
                i += 1
    except (OSError, IOError, FileNotFoundError) as e:  # Catch specific exceptions
        log.critical(f"** Unable to read the history file: {e}")
        return 1
    return 0


def export(args):
    """export environment variables"""
    # if command contains at least 1 space
    if args.count(" "):
        env = args.split(" ", 1)[1]
        # if it contains the equal sign, consider only the first one
        if env.count("="):
            var, value = env.split(" ")[0].split("=")[0:2]
            # disallow dangerous variable
            if var in variables.FORBIDDEN_ENVIRON:
                return 1, var
            # Strip the quotes from the value if it begins and ends with quotes (single or double)
            if (value.startswith('"') and value.endswith('"')) or (
                value.startswith("'") and value.endswith("'")
            ):
                value = value[1:-1]
            os.environ.update({var: value})
    return 0, None


def source(envfile):
    """Source a file in the current shell context"""
    envfile = os.path.expandvars(envfile)
    try:
        with open(envfile, encoding="utf-8") as env_vars:
            for env_var in env_vars.readlines():
                if env_var.split(" ", 1)[0] == "export":
                    export(env_var.strip())
    except (OSError, IOError):
        sys.stderr.write(f"ERROR: Unable to read environment file: {envfile}\n")
        return 1
    return 0


def cd(directory, conf):
    """implementation of the "cd" command"""
    # expand user's ~
    directory = os.path.expanduser(directory)

    # remove quotes if present
    directory = directory.strip("'").strip('"')

    if len(directory) >= 1:
        # add wildcard completion support to cd
        if directory.find("*"):
            # get all files and directories matching wildcard
            wildall = glob.glob(directory)
            wilddir = []
            # filter to only directories
            for item in wildall:
                if os.path.isdir(item):
                    wilddir.append(item)
            # sort results
            wilddir.sort()
            # if any results are returned, pick first one
            if len(wilddir) >= 1:
                directory = wilddir[0]
        # go previous directory
        if directory == "-":
            directory = conf["oldpwd"]

        # store current directory in oldpwd variable
        conf["oldpwd"] = os.getcwd()

        # change directory
        try:
            os.chdir(os.path.realpath(directory))
            conf["promptprint"] = utils.updateprompt(os.getcwd(), conf)
        except OSError as excp:
            sys.stdout.write(f"lshell: {directory}: {excp.strerror}\n")
            return excp.errno, conf
    else:
        os.chdir(conf["home_path"])
        conf["promptprint"] = utils.updateprompt(os.getcwd(), conf)

    return 0, conf


def check_background_jobs():
    """Check the status of background jobs and print a completion message if done."""
    global background_jobs
    updated_jobs = []
    for idx, job in enumerate(background_jobs, start=1):
        if job.poll() is None:
            # Process is still running
            updated_jobs.append((idx, job.args, job.pid))
        else:
            # Process has finished
            status = "Done" if job.returncode == 0 else "Failed"
            args = " ".join(job.args)
            # only print if the job has not been interrupted by the user
            if job.returncode != -2:
                print(f"[{idx}]+  {status}                    {args}")

            # Remove the job from the list of background jobs
            background_jobs.pop(idx - 1)


def get_job_status(job):
    """Return the status of a background job."""
    if job.poll() is None:
        status = "Stopped"
    elif job.poll() == 0:
        status = "Completed"  # Process completed successfully
    else:
        status = "Killed"  # Process was killed or terminated with a non-zero code
    return status


def jobs():
    """Return a list of background jobs."""
    global background_jobs
    joblist = []
    for idx, job in enumerate(background_jobs, start=1):
        status = get_job_status(job)
        if status in ["Stopped", "Killed"]:
            if job.poll() is not None:
                background_jobs.pop(idx - 1)
                continue
        cmd = " ".join(job.args)
        joblist.append([idx, status, cmd])
    return joblist


def print_jobs():
    """List all backgrounded jobs."""
    joblist = jobs()
    job_count = len(joblist)

    try:
        for i, job in enumerate(joblist, start=1):
            idx, status, cmd = job
            # Add '+' symbol for the most recent job
            job_symbol = "+"
            if job_count > 1:
                if i == job_count - 1:
                    # Add '-' for the second-to-last job
                    job_symbol = "-"
                elif i < job_count:
                    # No symbol for other jobs
                    job_symbol = " "
            print(f"[{idx}]{job_symbol}  {status}        {cmd}")
    except IndexError:
        return 1
    return 0


def bg_fg(job_type, job_id):
    """Resume a backgrounded job."""

    global background_jobs
    if job_type == "bg":
        print(f"lshell: bg not supported")
        return 1

    if job_id:
        # Check if job ID is valid
        try:
            job_id = int(job_id)
        except ValueError:
            print("Invalid job ID.")
            return 1
    else:
        # Use the last job if no specific job_id is provided
        if background_jobs:
            job_id = len(background_jobs)
        else:
            print(f"lshell: {job_type}: current: no such job")
            return 1

    if 0 < job_id <= len(background_jobs):
        job = background_jobs[job_id - 1]
        if job.poll() is None:
            if job_type == "fg":
                try:
                    print(" ".join(job.args))
                    # Bring it to the foreground and wait
                    os.killpg(os.getpgid(job.pid), signal.SIGCONT)
                    job.wait()
                    # Remove the job from the list if it has completed
                    if job.poll() is not None:
                        background_jobs.pop(job_id - 1)
                    return 0
                except KeyboardInterrupt:
                    os.killpg(os.getpgid(job.pid), signal.SIGINT)
                    background_jobs.pop(job_id - 1)
                    return 130
            # bg not supported at the moment
            # elif job_type == "bg":
            #     print(f"lshell: bg not supported")
            #     return 1
        else:
            print(f"lshell: {job_type}: {job_id}: no such job")
            return 1
    else:
        print(f"lshell: {job_type}: {job_id}: no such job")
        return 1



""" Setup script for lshell """

import os
import shutil
from setuptools import setup, find_packages
from setuptools.command.install import install

# import lshell specifics
from lshell.variables import __version__


class CustomInstallCommand(install):
    """Customized setuptools install command to handle etc files."""

    def run(self):
        # Call the standard install first
        install.run(self)

        # Determine correct configuration paths
        if os.geteuid() != 0:  # If not root, use ~/.local/etc
            etc_install_dir = os.path.join(os.path.expanduser("~"), ".local/etc")
        else:  # For system-wide install, use /etc
            etc_install_dir = "/etc"

        # Create necessary directories if they don't exist
        os.makedirs(os.path.join(etc_install_dir, "logrotate.d"), exist_ok=True)

        # Copy configuration files to appropriate directories
        shutil.copy("etc/lshell.conf", etc_install_dir)
        shutil.copy(
            "etc/logrotate.d/lshell", os.path.join(etc_install_dir, "logrotate.d")
        )


if __name__ == "__main__":

    with open("README.md", "r") as f:
        long_description = f.read()

    setup(
        name="limited-shell",
        version=__version__,
        description="lshell - Limited Shell",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="Ignace Mouzannar",
        author_email="ghantoos@ghantoos.org",
        maintainer="Ignace Mouzannar",
        maintainer_email="ghantoos@ghantoos.org",
        keywords=["limited", "shell", "security", "python"],
        url="https://github.com/ghantoos/lshell",
        project_urls={
            "GitHub": "https://github.com/ghantoos/lshell",
            "Changelog": "https://github.com/ghantoos/lshell/blob/master/CHANGELOG.md",
        },
        license="GPL-3",
        platforms=["UNIX"],
        scripts=["bin/lshell"],
        package_dir={"lshell": "lshell"},
        packages=find_packages(exclude=["test", "test.*"]),
        include_package_data=True,
        data_files=[
            ("etc", ["etc/lshell.conf"]),
            ("etc/logrotate.d", ["etc/logrotate.d/lshell"]),
            (
                "share/doc/lshell",
                ["README.md", "COPYING", "CHANGELOG.md", "SECURITY.md"],
            ),
            ("share/man/man1/", ["man/lshell.1"]),
        ],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Intended Audience :: Information Technology",
            "Intended Audience :: System Administrators",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: POSIX",
            "Programming Language :: Python :: 3",
            "Topic :: Security",
            "Topic :: System :: Shells",
            "Topic :: System :: System Shells",
            "Topic :: System :: Systems Administration",
            "Topic :: Terminals",
        ],
        python_requires=">=3.6",
        install_requires=[],
        cmdclass={
            "install": CustomInstallCommand,  # Use custom install command
        },
    )



"""Functional tests for lshell"""

import os
import unittest
import subprocess
from getpass import getuser
import tempfile
import shutil
import pexpect

# import lshell specifics
from lshell import utils

# pylint: disable=C0411
from test import test_utils

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"
LSHELL = f"{TOPDIR}/bin/lshell"
USER = getuser()
PROMPT = f"{USER}:~\\$"


class TestFunctions(unittest.TestCase):
    """Functional tests for lshell"""

    def setUp(self):
        """spawn lshell with pexpect and return the child"""
        self.child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --strict 1")
        self.child.expect(PROMPT)

    def tearDown(self):
        self.child.close()

    def do_exit(self, child):
        """Exit the shell"""
        child.sendline("exit")
        child.expect(pexpect.EOF)

    def test_01_welcome_message(self):
        """F01 | lshell welcome message"""
        expected = (
            "You are in a limited shell.\r\nType '?' or 'help' to get"
            " the list of allowed commands\r\n"
        )
        result = self.child.before.decode("utf8")
        self.assertEqual(expected, result)

    def test_02_builtin_ls_command(self):
        """F02 | built-in ls command"""
        p = subprocess.Popen(
            "ls ~", shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        cout = p.stdout
        expected = cout.read(-1)
        self.child.sendline("ls")
        self.child.expect(PROMPT)
        output = self.child.before.decode("utf8").split("ls\r", 1)[1]
        self.assertEqual(len(expected.strip().split()), len(output.strip().split()))

    def test_03_external_echo_command_num(self):
        """F03 | external echo number"""
        expected = "32"
        self.child.sendline("echo 32")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split()[2]
        self.assertEqual(expected, result)

    def test_04_external_echo_command_string(self):
        """F04 | external echo random string"""
        expected = "bla blabla  32 blibli! plop."
        self.child.sendline(f'echo "{expected}"')
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1].strip()
        self.assertEqual(expected, result)

    def test_05_external_echo_forbidden_syntax(self):
        """F05 | echo forbidden syntax $(bleh)"""
        expected = (
            '*** forbidden character -> "$("\r\n*** You '
            "have 1 warning(s) left, before getting kicked out.\r\nThis "
            "incident has been reported.\r\n"
        )
        self.child.sendline("echo $(uptime)")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_06_builtin_cd_change_dir(self):
        """F06 | built-in cd - change directory"""
        expected = ""
        home = os.path.expanduser("~")
        dirpath = None
        for path in os.listdir(home):
            dirpath = os.path.join(home, path)
            if os.path.isdir(dirpath):
                break
        if dirpath:
            self.child.sendline(f"cd {path}")
            self.child.expect(f"{USER}:~/{path}\\$")
            self.child.sendline("cd ..")
            self.child.expect(PROMPT)
            result = self.child.before.decode("utf8").split("\n", 1)[1]
            self.assertEqual(expected, result)

    def test_07_builtin_cd_tilda(self):
        """F07 | built-in cd - tilda bug"""
        expected = (
            '*** forbidden path -> "/etc/passwd"\r\n*** You have'
            " 1 warning(s) left, before getting kicked out.\r\nThis "
            "incident has been reported.\r\n"
        )
        self.child.sendline("ls ~/../../etc/passwd")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_08_builtin_cd_quotes(self):
        """F08 | built-in - quotes in cd "/" """
        expected = (
            '*** forbidden path -> "/"\r\n*** You have'
            " 1 warning(s) left, before getting kicked out.\r\nThis "
            "incident has been reported.\r\n"
        )
        self.child.sendline('ls -ld "/"')
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_09_external_forbidden_path(self):
        """F09 | external command forbidden path - ls /root"""
        expected = (
            '*** forbidden path -> "/root/"\r\n*** You have'
            " 1 warning(s) left, before getting kicked out.\r\nThis "
            "incident has been reported.\r\n"
        )
        self.child.sendline("ls ~root")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_10_builtin_cd_forbidden_path(self):
        """F10 | built-in command forbidden path - cd ~root"""
        expected = (
            '*** forbidden path -> "/root/"\r\n*** You have'
            " 1 warning(s) left, before getting kicked out.\r\nThis "
            "incident has been reported.\r\n"
        )
        self.child.sendline("cd ~root")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_11_etc_passwd_1(self):
        """F11 | /etc/passwd: empty variable 'ls "$a"/etc/passwd'"""
        if test_utils.is_alpine_linux():
            expected = "ls: $a/etc/passwd: No such file or directory\r\n"
        else:
            expected = (
                "ls: cannot access '$a/etc/passwd': No such file or directory\r\n"
            )
        self.child.sendline('ls "$a"/etc/passwd')
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_12_etc_passwd_2(self):
        """F12 | /etc/passwd: empty variable 'ls -l .*./.*./etc/passwd'"""
        if test_utils.is_alpine_linux():
            expected = "ls: .*./.*./etc/passwd: No such file or directory\r\n"
        else:
            expected = (
                "ls: cannot access '.*./.*./etc/passwd': No such file or directory\r\n"
            )
        self.child.sendline("ls -l .*./.*./etc/passwd")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_13a_etc_passwd_3(self):
        """F13(a) | /etc/passwd: empty variable 'ls -l .?/.?/etc/passwd'"""
        if test_utils.is_alpine_linux():
            expected = "ls: .?/.?/etc/passwd: No such file or directory\r\n"
        else:
            expected = (
                "ls: cannot access '.?/.?/etc/passwd': No such file or directory\r\n"
            )
        self.child.sendline("ls -l .?/.?/etc/passwd")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_13b_etc_passwd_4(self):
        """F13(b) | /etc/passwd: empty variable 'ls -l ../../etc/passwd'"""
        expected = (
            '*** forbidden path -> "/etc/passwd"\r\n*** You have'
            " 1 warning(s) left, before getting kicked out.\r\nThis "
            "incident has been reported.\r\n"
        )
        self.child.sendline("ls -l ../../etc/passwd")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_14_path_completion_tilda(self):
        """F14 | path completion with ~/"""
        p = subprocess.Popen(
            "ls -F ~/", shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        # Create two random files in the home directory
        file1 = os.path.join("random_file_1.txt")
        file2 = os.path.join("random_file_2.txt")
        with open(file1, "w") as f:
            f.write("This is a random file 1.")
        with open(file2, "w") as f:
            f.write("This is a random file 2.")
        cout = p.stdout
        expected = cout.read().decode("utf8").strip().split()
        self.child.sendline("cd \t\t")
        self.child.expect(PROMPT)
        output = (
            self.child.before.decode("utf8").strip().split("\n", 1)[1].strip().split()
        )
        output = [
            item
            for item in output
            if not item.startswith("--More--") and not item.startswith("\x1b")
        ]
        self.assertEqual(len(expected), len(output))

        # cleanup
        os.remove(file1)
        os.remove(file2)

    def test_15_cmd_completion_tab_tab(self):
        """F15 | command completion: tab to list commands"""
        expected = (
            "\x07\r\nbg       clear    exit     help     jobs     lpath    lsudo    "
            "\r\ncd       echo     fg       history  ll       ls"
        )
        self.child.sendline("\t\t")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").strip()

        self.assertEqual(expected, result)

    def test_16a_exitcode_with_separator_external_cmd(self):
        """F16(a) | external command exit codes with separator"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} " '--forbidden "[]"')
        child.expect(PROMPT)

        if test_utils.is_alpine_linux():
            expected_1 = "ls: nRVmmn8RGypVneYIp8HxyVAvaEaD55: No such file or directory"
        else:
            expected_1 = (
                "ls: cannot access 'nRVmmn8RGypVneYIp8HxyVAvaEaD55': "
                "No such file or directory"
            )
        expected_2 = "blabla"
        expected_3 = "0"
        child.sendline("ls nRVmmn8RGypVneYIp8HxyVAvaEaD55; echo blabla; echo $?")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")
        result_1 = result[1].strip()
        result_2 = result[2].strip()
        result_3 = result[3].strip()
        self.assertEqual(expected_1, result_1)
        self.assertEqual(expected_2, result_2)
        self.assertEqual(expected_3, result_3)
        self.do_exit(child)

    def test_16b_exitcode_with_separator_external_cmd(self):
        """F16(b) | external command exit codes with separator"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} " '--forbidden "[]"')
        child.expect(PROMPT)

        if test_utils.is_alpine_linux():
            expected_1 = "ls: nRVmmn8RGypVneYIp8HxyVAvaEaD55: No such file or directory"
            expected_2 = "1"
        else:
            expected_1 = (
                "ls: cannot access 'nRVmmn8RGypVneYIp8HxyVAvaEaD55': "
                "No such file or directory"
            )
            expected_2 = "2"
        child.sendline("ls nRVmmn8RGypVneYIp8HxyVAvaEaD55; echo $?")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")
        result_1 = result[1].strip()
        result_2 = result[2].strip()
        self.assertEqual(expected_1, result_1)
        self.assertEqual(expected_2, result_2)
        self.do_exit(child)

    def test_17_exitcode_without_separator_external_cmd(self):
        """F17 | external command exit codes without separator"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} " '--forbidden "[]"')
        child.expect(PROMPT)

        if test_utils.is_alpine_linux():
            expected = "1"
        else:
            expected = "2"
        child.sendline("ls nRVmmn8RGypVneYIp8HxyVAvaEaD55")
        child.expect(PROMPT)
        child.sendline("echo $?")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_18_cd_exitcode_with_separator_internal_cmd(self):
        """F18 | built-in command exit codes with separator"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} " '--forbidden "[]"')
        child.expect(PROMPT)

        expected = "2"
        child.sendline("cd nRVmmn8RGypVneYIp8HxyVAvaEaD55; echo $?")
        child.expect(PROMPT)
        child.sendline("echo $?")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_19_cd_exitcode_without_separator_external_cmd(self):
        """F19 | built-in exit codes without separator"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} " '--forbidden "[]"')
        child.expect(PROMPT)

        expected = "2"
        child.sendline("cd nRVmmn8RGypVneYIp8HxyVAvaEaD55")
        child.expect(PROMPT)
        child.sendline("echo $?")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_20_cd_with_cmd_unknwon_dir(self):
        """F20 | test built-in cd with command when dir does not exist
        Should be returning error, not executing cmd
        """
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} " '--forbidden "[]"')
        child.expect(PROMPT)

        expected = (
            "lshell: nRVmmn8RGypVneYIp8HxyVAvaEaD55: No such file or " "directory"
        )

        child.sendline("cd nRVmmn8RGypVneYIp8HxyVAvaEaD55; echo $?")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_21_allow_slash(self):
        """F21 | user should able to allow / access minus some directory
        (e.g. /var)
        """
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--path \"['/'] - ['/var']\""
        )
        child.expect(PROMPT)

        expected = "*** forbidden path: /var/"
        child.sendline("cd /")
        child.expect(f"{USER}:/\\$")
        child.sendline("cd var")
        child.expect(f"{USER}:/\\$")
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_22_expand_env_variables(self):
        """F22 | expanding of environment variables"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"+ ['export']\""
        )
        child.expect(PROMPT)

        expected = f"{os.path.expanduser('~')}/test"
        child.sendline("export A=test")
        child.expect(PROMPT)
        child.sendline("echo $HOME/$A")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_23_expand_env_variables_cd(self):
        """F23 | expanding of environment variables when using cd"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"+ ['export']\""
        )
        child.expect(PROMPT)

        random = utils.random_string(32)

        expected = f"lshell: {os.path.expanduser('~')}/random_{random}: No such file or directory"
        child.sendline(f"export A=random_{random}")
        child.expect(PROMPT)
        child.sendline("cd $HOME/$A")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_24_cd_and_command(self):
        """F24 | cd && command should not be interpreted by internal function"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG}")
        child.expect(PROMPT)

        expected = "OK"
        child.sendline('cd ~ && echo "OK"')
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_25_keyboard_interrupt(self):
        """F25 | test cat(1) with KeyboardInterrupt, should not exit"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"+ ['cat']\""
        )
        child.expect(PROMPT)

        child.sendline("cat")
        child.sendline(" foo ")
        child.sendcontrol("c")
        child.expect(PROMPT)
        try:
            result = child.before.decode("utf8").split("\n")[1].strip()
            # both behaviors are correct
            if result.startswith("foo"):
                expected = "foo"
            elif result.startswith("^C"):
                expected = "^C"
            else:
                expected = "unknown"
        except IndexError:
            # outputs u' ^C' on Debian
            expected = "^C"
            result = child.before.decode("utf8").strip()
        self.assertIn(expected, result)
        self.do_exit(child)

    def test_26_cmd_completion_dot_slash(self):
        """F26 | command completion: tab to list ./foo1 ./foo2"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"+ ['./foo1', './foo2']\""
        )
        child.expect(PROMPT)

        expected = "./\x07foo\x07\r\nfoo1  foo2"
        child.sendline("./\t\t\t")
        child.expect(PROMPT)
        result = child.before.decode("utf8").strip()

        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_27_checksecure_awk(self):
        """F27 | checksecure awk script with /bin/sh"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"+ ['awk']\""
        )
        child.expect(PROMPT)

        if test_utils.is_alpine_linux():
            command = "awk 'BEGIN {system(\"/bin/sh\")}'"
            expected = "*** forbidden path: /bin/busybox"
        else:
            command = "awk 'BEGIN {system(\"/usr/bin/bash\")}'"
            expected = "*** forbidden path: /usr/bin/bash"
        child.sendline(command)
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()

        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_28_catch_terminal_ctrl_j(self):
        """F28 | test ctrl-v ctrl-j then command, forbidden/security"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} ")
        child.expect(PROMPT)

        expected = "*** forbidden control char: echo\r"
        child.send("echo")
        child.sendcontrol("v")
        child.sendcontrol("j")
        child.sendline("bash")
        child.expect(PROMPT)

        result = child.before.decode("utf8").split("\n")

        self.assertIn(expected, result)
        self.do_exit(child)

    def test_29_catch_terminal_ctrl_k(self):
        """F29 | test ctrl-v ctrl-k then command, forbidden/security"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} ")
        child.expect(PROMPT)

        expected = "*** forbidden control char: echo\x0b() bash && echo\r"
        child.send("echo")
        child.sendcontrol("v")
        child.sendcontrol("k")
        child.sendline("() bash && echo")
        child.expect(PROMPT)

        result = child.before.decode("utf8").split("\n")[1]

        self.assertIn(expected, result)
        self.do_exit(child)

    def test_30_disable_exit(self):
        """F31 | test disabled exit command"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} " "--disable_exit 1 ")
        child.expect(PROMPT)

        expected = ""
        child.sendline("exit")
        child.expect(PROMPT)

        result = child.before.decode("utf8").split("\n")[1]

        self.assertIn(expected, result)

    def test_31_security_echo_freedom_and_help(self):
        """F31 | test help, then echo FREEDOM! && help () sh && help"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} ")
        child.expect(PROMPT)

        # Step 1: Enter `help` command
        expected_help_output = (
            "bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo"
        )
        child.sendline("help")
        child.expect(PROMPT)
        help_output = child.before.decode("utf8").split("\n", 1)[1].strip()

        self.assertEqual(expected_help_output, help_output)

        # Step 2: Enter `echo FREEDOM! && help () sh && help`
        expected_output = (
            "FREEDOM!\r\nbg  cd  clear  echo  exit  fg  help  history  "
            "jobs  ll  lpath  ls  lsudo\r\n"
            "bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo"
        )
        child.sendline("echo FREEDOM! && help () sh && help")
        child.expect(PROMPT)

        result = child.before.decode("utf8").strip().split("\n", 1)[1].strip()

        # Verify the combined output
        self.assertEqual(expected_output, result)
        self.do_exit(child)

    def test_32_security_echo_freedom_and_cd(self):
        """F32 | test echo FREEDOM! && cd () bash && cd ~/"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} ")
        child.expect(PROMPT)

        # Step 1: Enter `help` command
        expected_help_output = (
            "bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo"
        )
        child.sendline("help")
        child.expect(PROMPT)
        help_output = child.before.decode("utf8").split("\n", 1)[1].strip()

        self.assertEqual(expected_help_output, help_output)

        # Step 2: Enter `echo FREEDOM! && help () sh && help`
        expected_output = "FREEDOM!\r\nlshell: () bash: No such file or directory"
        child.sendline("echo FREEDOM! && cd () bash && cd ~/")
        child.expect(PROMPT)

        result = child.before.decode("utf8").strip().split("\n", 1)[1].strip()

        # Verify the combined output
        self.assertEqual(expected_output, result)
        self.do_exit(child)

    def test_33_ls_non_existing_directory_and_echo(self):
        """Test: ls non_existing_directory && echo nothing"""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG}")
        child.expect(PROMPT)

        child.sendline("ls non_existing_directory && echo nothing")
        child.expect(PROMPT)

        output = child.before.decode("utf8").split("\n", 1)[1].strip()
        # Since ls fails, echo nothing shouldn't run
        self.assertNotIn("nothing", output)
        self.do_exit(child)

    def test_34_ls_and_echo_ok(self):
        """Test: ls && echo OK"""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG}")
        child.expect(PROMPT)

        child.sendline("ls && echo OK")
        child.expect(PROMPT)

        output = child.before.decode("utf8").split("\n", 1)[1].strip()
        # ls succeeds, echo OK should run
        self.assertIn("OK", output)
        self.do_exit(child)

    def test_35_ls_non_existing_directory_or_echo_ok(self):
        """Test: ls non_existing_directory || echo OK"""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG}")
        child.expect(PROMPT)

        child.sendline("ls non_existing_directory || echo OK")
        child.expect(PROMPT)

        output = child.before.decode("utf8").split("\n", 1)[1].strip()
        # ls fails, echo OK should run
        self.assertIn("OK", output)
        self.do_exit(child)

    def test_36_ls_or_echo_nothing(self):
        """Test: ls || echo nothing"""
        child = pexpect.spawn(f"{LSHELL} --config {CONFIG}")
        child.expect(PROMPT)

        child.sendline("ls || echo nothing")
        child.expect(PROMPT)

        output = child.before.decode("utf8").split("\n", 1)[1].strip()
        # ls succeeds, echo nothing should not run
        self.assertNotIn("nothing", output)
        self.do_exit(child)

    def test_37_env_vars_file_not_found(self):
        """Test missing environment variable file"""
        missing_file_path = "/path/to/missing/file"

        # Inject the environment variable file path
        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} "
            f"--env_vars_files \"['{missing_file_path}']\""
        )

        # Expect the prompt after shell startup
        child.expect(PROMPT)

        # Simulate what happens when the environment variable file is missing
        expected = (
            f"ERROR: Unable to read environment file: {missing_file_path}\r\n"
            "You are in a limited shell.\r\n"
            "Type '?' or 'help' to get the list of allowed commands\r\n"
        )

        # Check the error message in the output
        self.assertIn(expected, child.before.decode("utf8"))
        self.do_exit(child)

    def test_38_load_env_vars_from_file(self):
        """Test loading environment variables from file"""

        # Create a temporary file to store environment variables
        with tempfile.NamedTemporaryFile(
            mode="w", delete=False, dir="/tmp"
        ) as temp_env_file:
            temp_env_file.write("export bar=helloworld\n")
            temp_env_file.flush()  # Ensure data is written to disk
            temp_env_file_path = temp_env_file.name

        # Set the temp env file path in the config
        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} "
            f"--env_vars_files \"['{temp_env_file_path}']\""
        )
        child.expect(PROMPT)

        # Test if the environment variable was loaded
        child.sendline("echo $bar")
        child.expect(PROMPT)

        result = child.before.decode("utf8").strip().split("\n", 1)[1].strip()
        self.assertEqual(result, "helloworld")

        # Cleanup the temporary file
        os.remove(temp_env_file_path)
        self.do_exit(child)

    def test_39_script_execution_with_template(self):
        """Test executing script after modifying shebang and clean up afterward"""

        template_path = f"{TOPDIR}/test/template.lsh"
        test_script_path = f"/tmp/test.lsh"

        # Step 1: Create the wrapper script
        with tempfile.NamedTemporaryFile(mode="w", delete=False, dir="/tmp") as wrapper:
            wrapper.write(
                f"""#!/bin/sh
exec {LSHELL} --config {CONFIG} "$@"
"""
            )
            wrapper.flush()  # Ensure data is written to disk
            wrapper_path = wrapper.name

        # Step 2: Copy template.lsh to test.lsh and replace the shebang
        shutil.copy(template_path, test_script_path)

        # Make the wrapper executable
        os.chmod(wrapper_path, 0o755)
        os.chmod(test_script_path, 0o755)

        # Replace the placeholder in the shebang
        with open(test_script_path, "r+") as f:
            content = f.read()
            content = content.replace("#!SHEBANG", f"#!{wrapper_path}")
            f.seek(0)
            f.write(content)
            f.truncate()

        # Spawn a child process to run the test.lsh script using pexpect
        child = pexpect.spawn(test_script_path)

        # Expected output
        expected_output = """test\r
*** forbidden command: dig\r
*** forbidden path: /tmp/\r
FREEDOM\r
bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo\r
bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo\r
*** forbidden path: /"""

        # Wait for the script to finish executing
        child.expect(pexpect.EOF)

        # Capture the output and compare with expected output
        result = child.before.decode("utf8").strip()
        self.assertEqual(result, expected_output)

        # Cleanup: remove the test script after the test
        if os.path.exists(test_script_path):
            os.remove(test_script_path)
        self.do_exit(child)

    def test_40_script_execution_with_template_strict(self):
        """Test executing script after modifying shebang and clean up afterward"""

        template_path = f"{TOPDIR}/test/template.lsh"
        test_script_path = f"/tmp/test.lsh"

        # Step 1: Create the wrapper script
        with tempfile.NamedTemporaryFile(mode="w", delete=False, dir="/tmp") as wrapper:
            wrapper.write(
                f"""#!/bin/sh
exec {LSHELL} --config {CONFIG} --strict 1 "$@"
"""
            )
            wrapper.flush()  # Ensure data is written to disk
            wrapper_path = wrapper.name

        # Step 2: Copy template.lsh to test.lsh and replace the shebang
        shutil.copy(template_path, test_script_path)

        # Make the wrapper executable
        os.chmod(wrapper_path, 0o755)
        os.chmod(test_script_path, 0o755)

        with open(test_script_path, "r+") as f:
            content = f.read()
            content = content.replace("#!SHEBANG", f"#!{wrapper_path}")
            f.seek(0)
            f.write(content)
            f.truncate()

        # Step 3: Spawn a child process to run the test.lsh script using pexpect
        child = pexpect.spawn(test_script_path)

        # Expected output
        expected_output = """test\r
*** forbidden command -> "dig"\r
*** You have 1 warning(s) left, before getting kicked out.\r
This incident has been reported.\r
*** forbidden path -> "/tmp/"\r
*** You have 0 warning(s) left, before getting kicked out.\r
This incident has been reported.\r
FREEDOM\r
bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo\r
bg  cd  clear  echo  exit  fg  help  history  jobs  ll  lpath  ls  lsudo\r
*** forbidden path -> "/"\r
*** Kicked out"""

        # Wait for the script to finish executing
        child.expect(pexpect.EOF)

        # Capture the output and compare with expected output
        result = child.before.decode("utf8").strip()
        self.assertEqual(result, expected_output)

        # Step 5: Cleanup: remove the test script and wrapper after the test
        if os.path.exists(test_script_path):
            os.remove(test_script_path)
        if os.path.exists(wrapper_path):
            os.remove(wrapper_path)
        self.do_exit(child)

    def test_41_multicmd_with_wrong_arg_should_fail(self):
        """F20 | Allowing 'echo asd': Test 'echo qwe' should fail"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"['echo asd']\""
        )
        child.expect(PROMPT)

        expected = "*** forbidden command: echo"

        child.sendline("echo qwe")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_42_multicmd_with_near_exact_arg_should_fail(self):
        """F41 | Allowing 'echo asd': Test 'echo asds' should fail"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"['echo asd']\""
        )
        child.expect(PROMPT)

        expected = "*** forbidden command: echo"

        child.sendline("echo asds")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_43_multicmd_without_arg_should_fail(self):
        """F42 | Allowing 'echo asd': Test 'echo' should fail"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"['echo asd']\""
        )
        child.expect(PROMPT)

        expected = "*** forbidden command: echo"

        child.sendline("echo")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_44_multicmd_asd_should_pass(self):
        """F43 | Allowing 'echo asd': Test 'echo asd' should pass"""

        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--allowed \"['echo asd']\""
        )
        child.expect(PROMPT)

        expected = "asd"

        child.sendline("echo asd")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n")[1].strip()
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_45_overssh_allowed_command_exit_0(self):
        """F44 | Test 'ssh -c ls' command should exit 0"""
        # add SSH_CLIENT to environment
        if not os.environ.get("SSH_CLIENT"):
            os.environ["SSH_CLIENT"] = "random"

        self.child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " f"--overssh \"['ls']\" " f"-c 'ls'"
        )
        self.child.expect(pexpect.EOF, timeout=10)

        # Assert that the process exited
        self.assertIsNotNone(
            self.child.exitstatus,
            f"The lshell process did not exit as expected: {self.child.exitstatus}",
        )

        # Optionally, you can assert that the exit code is correct
        self.assertEqual(
            self.child.exitstatus,
            0,
            f"The process should exit with code 0, got {self.child.exitstatus}.",
        )

    def test_46_overssh_allowed_command_exit_1(self):
        """F44 | Test 'ssh -c ls' command should exit 1"""
        # add SSH_CLIENT to environment
        if not os.environ.get("SSH_CLIENT"):
            os.environ["SSH_CLIENT"] = "random"

        self.child = pexpect.spawn(
            f"{LSHELL} "
            f"--config {CONFIG} "
            f"--overssh \"['ls']\" "
            f"-c 'ls /random'"
        )
        self.child.expect(pexpect.EOF, timeout=10)

        # Assert that the process exited
        self.assertIsNotNone(
            self.child.exitstatus, "The lshell process did not exit as expected."
        )

        # Optionally, you can assert that the exit code is correct
        self.assertEqual(
            self.child.exitstatus,
            1,
            f"The process should exit with code 1, got {self.child.exitstatus}.",
        )

    def test_46_overssh_not_allowed_command_exit_1(self):
        """F44 | Test 'ssh -c lss' command should succeed"""
        # add SSH_CLIENT to environment
        if not os.environ.get("SSH_CLIENT"):
            os.environ["SSH_CLIENT"] = "random"

        self.child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " f"--overssh \"['ls']\" " f"-c 'lss'"
        )
        self.child.expect(pexpect.EOF, timeout=10)

        # Assert that the process exited
        self.assertIsNotNone(
            self.child.exitstatus, "The lshell process did not exit as expected."
        )

        # Optionally, you can assert that the exit code is correct
        self.assertEqual(
            self.child.exitstatus,
            1,
            f"The process should exit with code 1, got {self.child.exitstatus}.",
        )

    def test_47_backticks(self):
        """F47 | Forbidden backticks should be reported"""
        expected = (
            '*** forbidden character -> "`"\r\n'
            "*** You have 1 warning(s) left, before getting kicked out.\r\n"
            "This incident has been reported.\r\n"
        )
        self.child.sendline("echo `uptime`")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_48_replace_backticks_with_dollar_parentheses(self):
        """F48 | Forbidden syntax $(command) should be reported"""
        expected = (
            '*** forbidden character -> "$("\r\n'
            "*** You have 1 warning(s) left, before getting kicked out.\r\n"
            "This incident has been reported.\r\n"
        )
        self.child.sendline("echo $(uptime)")
        self.child.expect(PROMPT)
        result = self.child.before.decode("utf8").split("\n", 1)[1]
        self.assertEqual(expected, result)

    def test_49_env_variable_with_dollar_braces(self):
        """F49 | Syntax ${command} should replace with the variable"""
        child = pexpect.spawn(
            f"{LSHELL} " f"--config {CONFIG} " "--env_vars \"{'foo':'OK'}\""
        )
        child.expect(PROMPT)

        child.sendline("echo ${foo}")
        child.expect(PROMPT)
        result = child.before.decode("utf8").split("\n", 1)[1]
        expected = "OK\r\n"
        self.assertEqual(expected, result)
        self.do_exit(child)

    def test_50_warnings_then_kickout(self):
        """F50 | kicked out after warning counter"""
        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} --strict 1 --warning_counter 0"
        )
        child.sendline("lslsls")
        child.sendline("lslsls")
        child.expect(pexpect.EOF, timeout=10)

        # Assert that the process exited
        self.assertIsNotNone(
            child.exitstatus, "The lshell process did not exit as expected."
        )

        # Optionally, you can assert that the exit code is correct
        self.assertEqual(child.exitstatus, 1, "The process should exit with code 1.")
        self.do_exit(child)
